from collections import deque
from lxml.html import HtmlElement


class ListExtractor:
    def extract(self, element: HtmlElement, feature):
        result = []
        if feature.startswith('/'):
            feature_element = element.xpath(feature)
        else:
            feature_element = element.xpath(f'//*[contains(text(), "{feature}")]')

        if not feature_element:
            print('找不到 feature！')
            return result

        parent = feature_element[0]
        leaf_class = parent.attrib.get('class', '')
        if leaf_class:
            leaf_node = f'{parent.tag}[@class="{leaf_class}"]'
        else:
            leaf_node = parent.tag
        is_a_tag = parent.tag == 'a'
        sub_path_queue = deque([leaf_node])
        while parent is not None:
            parent = parent.getparent()
            if parent is None:
                break
            path = '/'.join(sub_path_queue)
            item_list = parent.xpath(path)
            if len(item_list) > 3:
                for item in item_list:
                    item_info = {'title': ''.join(item.xpath('text()'))}
                    if is_a_tag:
                        item_info['url'] = ''.join(item.xpath('@href'))
                    result.append(item_info)
                return result
            sub_path_queue.insert(0, parent.tag)
        return result





import json
import re
from datetime import datetime

import jsonpath
from parsel import Selector

from ..dom.DomHandler import DomHandle
from ..utils import config
from lxml.html import HtmlElement, fromstring, tostring, HTMLParser, document_fromstring
from dateutil.parser import parse
from ..defaults import DATETIME_PATTERN, PUBLISH_TIME_META, RE_SUB_MAP, SCRIPT_TIME_RE, JW_DATETIME_PATTERN, \
    PUBLISH_TIME_TAG, DATETIME_PATTERN_COMPARE, TIME_TOKEN, SCRIPT_TIME_RE_TOKEN, BASE_SCRIPT_TIME_RE


class TimeExtractor:
    def __init__(self):
        self.time_pattern = JW_DATETIME_PATTERN + DATETIME_PATTERN
        self.dom_handler = DomHandle()
        self.SCRIPT_TIME_RE = SCRIPT_TIME_RE

    def extractor(self, element: HtmlElement, publish_time_xpath: str = '', normal_html=None, title=None,
                  content=None, html=None, npp_pt=None) -> dict:
        '''
        时间解析器
        :param element: 网页对象
        :param publish_time_xpath: 发布时间xpath
        :param normal_html: 使用 NFKC 对网页源代码进行归一化，把特殊符号转换为普通符号
        :param title: 标题
        :param content: 内容
        :param html: 原始html源码
        :param text: 暂无定义使用
        :return:
        /*
                发布时间会整理以下四种格式以字典的形式返回
                {
                    'publish_time_src': '2012-12-10T09:24:01+00:00', # 发布时间原始数据
                    'publish_time_ts': 1355131441, # 发布时间时间戳格式
                    'publish_time_zone': UTC, # 发布时间的时区
                    'publish_time_format': '2012-12-10 09:24:01' # 发布时间格式化后的数据
                }
        */
        '''

        # normal_html = normal_html,
        # title = title,
        # content = content[0][1]['text'],
        # html = html,
        # text = None
        publish_time_xpath = publish_time_xpath or config.get('publish_time', {}).get('xpath')
        publish_time = (self.extract_from_user_xpath(publish_time_xpath, element)  # 用户指定的 Xpath 是第一优先级
                        or self.extract_from_meta(element)  # 第二优先级从 Meta 中提取
                        or npp_pt  # 引入3k的发布时间判断
                        or self.extract_data_from_ld_script(html_text=normal_html)  # 第三优先级从javascript中有Json的情况
                        or self.extract_data_from_script_re(html_text=normal_html)  # 第三优先级从javascript中与正则情况
                        or self.find_data_in_all_html_text(normal_html)  # 第三优先级从全文正则情况
                        or self.extract_time_area(title=title, element=element,
                                                  content=content)  # 第四优先级从标题字符串和内容元素定位时间区域，提取出其文本信息并返回
                        or self.extract_from_target_xpath(html)  # 第五优先级函数根据预定义的 XPath 规则提取发布时间信息
                        or self.extract_from_text(element)  # 第六优先级首先对文本进行预处理，然后尝试使用预定义的时间匹配模式进行提取
                        or self.extract_from_full_text(html=normal_html)  # 最后全文检索
                        )

        return self.deal_publish_time(publish_time)

    def extract_from_user_xpath(self, publish_time_xpath: str, element: HtmlElement) -> str:
        if publish_time_xpath:
            publish_time = ''.join(element.xpath(publish_time_xpath))
            return publish_time
        return ''

    def is_have_at_least_one_digit(self, string: str) -> bool:
        lst = []
        for char in string:
            lst.append(str(char).isdigit())
        return any(lst)

    def extract_from_meta(self, element: HtmlElement) -> str:
        """
        从 META 数据中提取时间信息。

        参数：
        - element：网页源代码对应的 DOM 树。

        返回值：
        - str：提取到的时间信息。

        注意：
        函数首先检查预定义的 XPath 表达式列表 PUBLISH_TIME_META，尝试从网页的 META 标签中提取时间信息。
        如果找到符合条件的时间信息，立即返回。
        如果未找到符合条件的时间信息，则继续检查所有 META 标签，尝试从中提取包含时间相关信息的标签内容。
        """

        # 遍历预定义的 XPath 表达式列表 PUBLISH_TIME_META
        for xpath in PUBLISH_TIME_META:
            # 使用 XPath 表达式从 DOM 树中提取元素
            publish_time = element.xpath(xpath)
            # 如果找到发布时间并且至少包含一个数字，则返回时间信息
            if publish_time and self.is_have_at_least_one_digit(publish_time[0]):
                return '|'.join([i for i in publish_time if i])  # 返回找到的时间信息
            # 如果找到发布时间，则返回时间信息
            if publish_time:
                return ''.join(publish_time)  # 返回找到的时间信息

        # 如果未在 META 数据中找到时间信息，则继续检查所有 META 标签
        try:
            meta_tags = element.xpath("//meta")
            for meta_tag in meta_tags:
                name = meta_tag.get("name", "").lower()
                content = meta_tag.get("content", "").strip()
                # 如果 META 标签的 name 属性包含 'date'、'time' 或 'create'，则返回标签内容作为时间信息
                if 'date' in name or 'time' in name or 'create' in name:
                    return content  # 返回找到的时间信息
        except Exception as e:
            pass

        return ''  # 如果未找到时间信息，则返回空字符串

    def extract_from_target_xpath(self, html_str) -> str:
        """
        从给定的 HTML 字符串中提取发布时间信息。

        参数：
        - html_str：包含目标信息的 HTML 字符串。

        返回值：
        - str：提取到的发布时间信息。

        注意：
        函数根据预定义的 XPath 规则提取发布时间信息。
        """

        # 使用 lxml 解析 HTML 字符串并构建元素树
        element = fromstring(html_str)

        # 遍历预定义的 XPath 规则列表
        for xpath in PUBLISH_TIME_TAG:
            publish_time = element.xpath(xpath)
            # 如果找到发布时间信息，返回第一个匹配的结果（去除首尾空白）
            if publish_time:
                if len(publish_time) > 1:
                    new_xpath = xpath + '[1]'
                    publish_time = element.xpath(new_xpath)
                    return str(publish_time[0]).strip()
                else:
                    return str(publish_time[0]).strip()

        # 如果未找到发布时间信息，返回空字符串
        return ''

    def extract_from_full_text(self, html: str):
        """
        从给定的 HTML 文本中提取时间信息。

        参数：
        - html：要提取时间的 HTML 文本。

        返回值：
        - str：提取到的时间信息。

        注意：
        函数首先遍历 HTML 元素，检查元素的属性值和文本内容是否包含可能的时间信息。
        如果找到符合条件的时间信息，立即返回。
        如果未找到符合条件的时间信息，则重新遍历 HTML 元素的文本内容，并使用预定义的时间匹配模式进行提取。
        """

        dt_obj = ''  # 用于存储找到的时间对象

        root_element = fromstring(html)

        # 遍历 HTML 元素
        for element in root_element.iter():
            # 检查元素的属性值
            for key, value in element.attrib.items():
                # 如果属性值长度在5到30之间，可能包含时间信息
                if 5 < len(value.strip()) < 30:
                    # 尝试匹配预定义的时间正则表达式
                    for pattern in SCRIPT_TIME_RE:
                        result = re.search(pattern=pattern, string=value)
                        if result:
                            return result.group(1)  # 返回匹配到的时间信息

            # 检查元素的文本内容
            text = element.text
            if text:
                try:
                    if 5 < len(text.strip()) < 30:
                        # 尝试匹配预定义的时间正则表达式
                        for dt in self.time_pattern:
                            dt_obj = re.search(dt, text, flags=re.I)
                            if dt_obj:
                                return dt_obj.group(1)  # 返回匹配到的时间信息
                except Exception as e:
                    return ''

        # 如果未找到符合条件的时间信息，则重新遍历 HTML 元素的文本内容
        if not dt_obj:
            text = ''
            for element in root_element.iter():
                text += ''.join(element.xpath('.//text()'))  # 获取 HTML 元素的文本内容

            # 根据预定义的替换映射表对文本内容进行处理
            for pattern, repl_value in RE_SUB_MAP.items():
                text = re.sub(pattern, repl_value, text, flags=re.I)

            # 尝试匹配预定义的时间正则表达式
            for dt in self.time_pattern:
                dt_obj = re.search(dt, text, flags=re.I)
                if dt_obj:
                    if '1970' not in dt_obj.group():
                        return dt_obj.group()  # 返回匹配到的时间信息
                    else:
                        continue

        return ''  # 如果未找到时间信息，则返回空字符串

    def check_time(self, time_list, text):
        text = '        ' + text
        if time_list and isinstance(time_list, list):
            if len(time_list) > 0:
                for matched_time in time_list:
                    res = re.search(f'(........){matched_time}', text, flags=re.I)
                    if res:
                        time_prefix = res.group(1)
                        if [i for i in TIME_TOKEN if i.lower() in time_prefix.lower()]:
                            return matched_time
                        text = text.replace(matched_time, '', 1)
        return None

    def extract_from_text(self, element: HtmlElement = None, text=None) -> str:
        """
        从给定的文本中提取时间信息。

        参数：
        - element：要提取时间的 HTML 元素对象（可选）。
        - text：要提取时间的文本内容（可选）。

        返回值：
        - str：提取到的时间信息。

        注意：
        函数首先对文本进行预处理，然后尝试使用预定义的时间匹配模式进行提取。
        如果提取成功，返回匹配到的第一个时间信息；否则返回空字符串。
        """

        # 如果未提供文本，则从给定元素中获取文本内容
        if text is None:
            text = self.dom_handler.get_text(element=element, seg=' ', remove_noise=True, body=True)

        # 预处理文本，替换一些常见的时间相关词语
        text = re.sub(r'下午|上午|星期.?', ' ', text)
        text = re.sub(r'\s+', ' ', text)

        # 尝试使用预定义的时间匹配模式进行提取
        for dt in self.time_pattern:
            dt_obj = re.search(dt, text)
            if dt_obj:
                return dt_obj.group(1)

        # 根据预定义的替换映射表对文本进行进一步处理
        for pattern, repl_value in RE_SUB_MAP.items():
            text = re.sub(pattern, repl_value, text, flags=re.I)

        # 尝试使用更复杂的时间匹配模式进行提取
        for dt in DATETIME_PATTERN_COMPARE:
            time_list = re.findall(dt, text, flags=re.S)
            check_result = self.check_time(time_list, text)
            if check_result is not None:
                return check_result

        # 如果未找到时间信息，尝试提取第一个匹配的时间信息
        for dt in self.time_pattern:
            dt_obj = re.search(dt, text, flags=re.I)
            if dt_obj:
                return dt_obj.group(1)

        # 如果仍未找到时间信息，返回空字符串
        return ''

    def extract_time_area(self, title, element, content) -> str:
        """
        从给定的元素中提取时间区域文本。

        参数：
        - title：标题字符串，用于定位时间区域。
        - element：HtmlElement 对象，包含了需要提取时间区域的元素。
        - content：HtmlElement 对象，表示整个文档的内容。

        返回值：
        - str：提取到的时间区域文本。

        注意：
        函数将根据标题字符串和内容元素定位时间区域，提取出其文本信息并返回。
        """

        # 确保 element 是 HtmlElement 类型且 content 不为空
        if isinstance(element, HtmlElement) and content:
            # 获取内容元素的文本信息
            content_text = self.dom_handler.get_text(content)
            # 如果标题字符串出现在内容文本中，返回空字符串
            if title in content_text:
                return ''
            else:
                # 从内容元素中提取纯文本信息
                body_text = self.dom_handler.get_text(element=element, remove_noise=True, body=True)
                # 获取内容文本的前 10 个字符作为检测标记
                news_start = content_text[:10] if len(content_text) > 9 else content_text
                # 如果标记不在提取的文本中，重新解析元素并提取文本信息
                if news_start not in body_text:
                    new_html = bytes.decode(tostring(element, encoding='utf-8', pretty_print=False, method='html'))
                    utf8_parser = HTMLParser(encoding='utf-8')
                    element = document_fromstring(new_html.encode('utf-8', 'replace'), parser=utf8_parser)
                    body_text = self.dom_handler.get_text(element=element, remove_noise=True, body=True)
                # 替换特殊字符
                news_start = news_start.replace('\xa0', ' ')
                body_text.replace('\xa0', ' ')
                try:
                    # 查找标题字符串在文本中的位置
                    begin = body_text.index(title) + len(title)
                    # 如果标记在文本中，确定时间区域的结束位置
                    if news_start in body_text:
                        end = body_text[begin:].rindex(news_start) + begin
                        if end - begin > 300:
                            end = begin + 300
                    else:
                        end = begin + 300
                    # 如果开始位置大于等于结束位置，返回空字符串
                    if begin >= end:
                        time_area = ''
                    else:
                        # 提取时间区域文本
                        time_area = body_text[begin:end + 1]
                except:
                    time_area = ''
                # 如果无法提取时间区域，尝试提取标题之前的文本作为时间区域
                if not time_area:
                    try:
                        end = body_text.rindex(news_start)
                        time_area = body_text[:end + 1]
                    except:
                        return ''
                # 如果提取到时间区域文本，继续提取其中的信息并返回
                else:
                    return self.extract_from_text(None, text=time_area)
        return ''

    def extract_from_target_re(self, html, class_name, ):
        sel = Selector(text=html)
        scripts = sel.xpath('//script').getall()
        new_re_pattern = BASE_SCRIPT_TIME_RE % class_name
        return self.find_date(scripts, [new_re_pattern])

    def find_date(self, scripts, patterns):
        for script_source_code in scripts:
            for re_pattern in patterns:
                result = re.search(pattern=re_pattern, string=script_source_code)
                if result:
                    return result.group(1)
        return ''

    def find_data_in_all_html_text(self, text):
        """
        在 HTML 文本中查找时间信息。

        参数：
        - text：HTML 文本字符串。

        返回值：
        - str：找到的时间信息。

        注意：
        函数依次使用预定义的正则表达式列表 SCRIPT_TIME_RE 在文本中查找时间信息。
        如果找到符合条件的时间信息，则立即返回。
        """

        # 依次使用预定义的正则表达式列表在文本中查找时间信息
        for re_pattern in self.SCRIPT_TIME_RE:
            result = re.search(pattern=re_pattern, string=text)
            if result:
                return result.group(1)  # 如果找到符合条件的时间信息，则立即返回

        # 如果未找到时间信息，则返回空字符串
        return ''

    def extract_data_from_ld_script(self, html_text):
        """
        从包含在 <script type="application/ld+json"> 标签中的 JSON 数据中提取时间信息。

        参数：
        - html_text：HTML 文本字符串。

        返回值：
        - str：提取到的时间信息。

        注意：
        函数首先使用 XPath 表达式 '//script[@type="application/ld+json"]/text()' 提取 HTML 文本中的所有 JSON 数据。
        然后，遍历每个 JSON 数据，尝试从中提取时间信息。
        如果找到符合条件的时间信息，则立即返回。
        """

        # 使用 Scrapy 的 Selector 对象从 HTML 文本中提取 JSON 数据
        sel = Selector(text=html_text)
        datas = sel.xpath('//script[@type="application/ld+json"]/text()').getall()
        published = ''
        for script_content in datas:
            try:
                data = json.loads(script_content)
                # 遍历预定义的时间字段列表
                for key in SCRIPT_TIME_RE_TOKEN:
                    try:
                        published = jsonpath.jsonpath(data, f'$..{key}')
                        # 如果找到时间信息，则返回该信息
                        if published:
                            published = published[0]
                    except Exception:
                        pass
            except Exception:
                pass
        return published  # 如果未找到时间信息，则返回空字符串

    def extract_data_from_script_re(self, html_text):
        """
        从 HTML 文本中的 <script> 标签中提取时间信息。

        参数：
        - html_text：HTML 文本字符串。

        返回值：
        - str：提取到的时间信息。

        注意：
        函数首先使用 Scrapy 的 Selector 对象从 HTML 文本中提取所有的 <script> 标签。
        然后，调用 find_date 函数，使用预定义的正则表达式列表 SCRIPT_TIME_RE 在每个 <script> 标签中查找时间信息。
        如果找到符合条件的时间信息，则立即返回。
        """

        # 使用 Scrapy 的 Selector 对象从 HTML 文本中提取所有的 <script> 标签
        sel = Selector(text=html_text)
        scripts = sel.xpath('//script').getall()

        # 调用 find_date 函数，在每个 <script> 标签中查找时间信息
        res = self.find_date(scripts, self.SCRIPT_TIME_RE)

        return res  # 返回找到的时间信息

    def extract_from_script(self, html: str):

        res = self.extract_data_from_ld_script(html_text=html) or self.extract_data_from_script_re(
            html_text=html) or self.find_data_in_all_html_text(html)

        return res

    def deal_publish_time(self, publish_time_temp) -> dict:
        try:
            publish_time = parse(publish_time_temp)
            # 将字符串转换为 datetime 对象
            input_time = datetime.fromisoformat(publish_time.isoformat())
            # 获取时区信息
            timezone_info = input_time.tzname()
            if publish_time:
                formatted_time = {
                    'publish_time_src': publish_time_temp,
                    'publish_time_ts': int(publish_time.timestamp()),
                    'publish_time_zone': timezone_info,
                    'publish_time_format': publish_time.strftime('%Y-%m-%d %H:%M:%S')
                }
                return formatted_time
            else:
                return self._set_default_pt()
        except Exception as e:
            return self._set_default_pt()

    def deal_publish_time_dtt(self, publish_time_temp) -> dict:
        '''
        处理时间格式为日期时间型
        '''
        try:
            # 将字符串转换为 datetime 对象
            input_time = datetime.fromisoformat(publish_time_temp.isoformat())
            publish_time_temp2 = publish_time_temp.strftime('%Y-%m-%d %H:%M:%S')
            publish_time = parse(publish_time_temp2)
            # 获取时区信息
            timezone_info = input_time.tzname()
            if publish_time:
                formatted_time = {
                    'publish_time_src': publish_time_temp2,
                    'publish_time_ts': int(input_time.timestamp()),
                    'publish_time_zone': timezone_info,
                    'publish_time_format': publish_time.strftime('%Y-%m-%d %H:%M:%S')
                }
                return formatted_time
            else:
                return self._set_default_pt()
        except Exception as e:
            return self._set_default_pt()

    def _set_default_pt(self):
        '''
        设置默认的发布时间
        :return:
        '''
        # cts = int(time.time())
        # ctime = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(cts))
        # publish_time = parse(ctime)
        return {
            'publish_time_src': '',
            'publish_time_ts': '',
            'publish_time_zone': '',
            'publish_time_format': ''
        }



import re
import locale
import langid
from lxml import html


class LangExtractor:
    def __init__(self):
        self.name = "langid"
        # 用于提取语言代码的正则表达式模式
        self.langcode_pattern = re.compile(r'\b[a-zA-Z]{2}(?=([-_]|\b))')

    def language(self, response):
        try:
            root = html.fromstring(response)
        except ValueError:
            root = html.fromstring(response.encode("utf-8"))

        # 从不同的源提取语言
        lang = self.extract_from_attributes(root) or \
               self.extract_from_meta_tags(root) or \
               self.extract_from_articles(root) or \
               self.extract_from_body(root)

        if lang:
            lang = self.normalize_language(lang)

        return lang

    def extract_from_attributes(self, root):
        # 从 HTML 属性中提取语言代码
        lang = root.get('lang') or root.get('xml:lang')
        return lang

    def extract_from_meta_tags(self, root):
        # 从 meta 标签中提取语言代码
        lang = root.xpath('string(//meta[@name="language"]/@content)') or \
               root.xpath('string(//meta[@property="og:locale"]/@content)')
        return lang.strip() if lang else None

    def extract_from_articles(self, root):
        # 从文章中提取语言代码，选择最长的文章
        article_texts = [re.sub(r'\s+', ' ', article.text_content().strip()) for article in root.xpath('//article')]
        longest_article = max(article_texts, key=len, default='')
        return langid.classify(longest_article)[0] if longest_article else None

    def extract_from_body(self, root):
        # 从整个 HTML 主体中提取语言代码
        return langid.classify(root.text_content().strip())[0]

    def normalize_language(self, lang):
        # 规范化语言代码的输出
        matches = self.langcode_pattern.search(lang)
        if matches:
            return matches.group(0)
        else:
            normalized = locale.normalize(re.split(r'\s|;|,', lang.strip())[0])
            matches = self.langcode_pattern.search(normalized)
            return matches.group(0) if matches else None



from lxml.etree import tostring

from .utils import pre_parse, remove_noise_node, config, html2element, normalize_text
from .extractor import ContentExtractor, TitleExtractor, TimeExtractor, AuthorExtractor, ListExtractor, LangExtractor, \
    HeadMetaExtractor
from newspaper import Config, Article


class GeneralNewsExtractor:
    def __init__(self):
        # 配置 newspaper
        self.config = Config()
        self.config.fetch_images = False

    def extract(self,
                html,
                title_xpath='',
                author_xpath='',
                publish_time_xpath='',
                host='',
                body_xpath='',
                noise_node_list=None,
                with_body_html=False,
                use_visiable_info=False):
        article = Article('', config=self.config)
        try:
            # 对 HTML 进行预处理可能会破坏 HTML 原有的结构，导致根据原始 HTML 编写的 XPath 不可用
            # 因此，如果指定了 title_xpath/author_xpath/publish_time_xpath，那么需要先提取再进行
            article.set_html(html)
            article.parse()
            # 输出已解析的信息
            info_dict = {
                'movies': article.movies,
                'keywords': article.keywords,
                'meta_keywords': article.meta_keywords,
                'meta_description': article.meta_description,
                'summary': article.summary,
                'tags': article.tags or [],
                'canonical_link': article.canonical_link,
                'source_url': article.source_url,
            }
            body_html = ''
            if article.clean_top_node:
                body_html = tostring(article.clean_top_node, encoding="unicode", method="html")
            try:
                # 预处理
                normal_html = normalize_text(html)
                element = html2element(normal_html)
                lang = LangExtractor().language(html)
                headmeta = HeadMetaExtractor().extractor(element)
                title = TitleExtractor().extract(element, title_xpath=title_xpath)
                author = AuthorExtractor().extractor(element, author_xpath=author_xpath) or article.authors
                element = pre_parse(element)
                remove_noise_node(element, noise_node_list)
                if not host:
                    host = headmeta.get('host')
                content = ContentExtractor().extract(element,
                                                     host=host,
                                                     with_body_html=with_body_html,
                                                     body_xpath=body_xpath,
                                                     use_visiable_info=use_visiable_info)
                publish_time = TimeExtractor().extractor(element,
                                                         publish_time_xpath=publish_time_xpath,
                                                         normal_html=normal_html,
                                                         title=title,
                                                         content=content[0][1]['text'],
                                                         html=html,
                                                         npp_pt=article.publish_date
                                                         )
                text = content[0][1]['text']
                npp_text = article.text
                if npp_text and len(npp_text) > len(text):
                    text = npp_text
                else:
                    try:
                        body_html = content[0][1].get('body_html', '')
                    except:
                        pass
                result = {'title': title,
                          'author': author,
                          'publish_time': publish_time,
                          'lang': lang,
                          'content': text,
                          'images': content[0][1]['images'],
                          'headmeta': headmeta,
                          'top_image': headmeta.pop('top_image'),
                          'website': headmeta.pop('host')
                          }
            except:
                publish_time_temp = article.publish_date
                publish_time = TimeExtractor().deal_publish_time_dtt(publish_time_temp)
                result = {
                    'title': article.title,
                    'author': article.authors,
                    'publish_time': publish_time,
                    'lang': article.meta_lang,
                    'content': article.text,
                    'images': article.imgs,
                    'headmeta': dict(article.meta_data),
                    'top_image': article.top_image,
                    'website': host
                }
            if with_body_html or config.get('with_body_html', False):
                result['body_html'] = body_html
            result = dict(result, **info_dict)
            return result
        except Exception as e:
            return {'error': "parser error," + str(e)}


class ListPageExtractor:
    def extract(self, html, feature):
        normalize_html = normalize_text(html)
        element = html2element(normalize_html)
        extractor = ListExtractor()
        return extractor.extract(element, feature)



import codecs
import logging
import os
import re
import yaml
import unicodedata
from lxml.html import fromstring, HtmlElement
from lxml.html import etree
from urllib.parse import urlparse, urljoin
from http.cookiejar import CookieJar as cj

from .version import __version__

from .defaults import USELESS_TAG, TAGS_CAN_BE_REMOVE_IF_EMPTY, USELESS_ATTR, HIGH_WEIGHT_ARRT_KEYWORD, ALLOWED_TYPES

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def normalize_node(element: HtmlElement):
    etree.strip_elements(element, *USELESS_TAG)
    for node in iter_node(element):
        # inspired by readability.
        if node.tag.lower() in TAGS_CAN_BE_REMOVE_IF_EMPTY and is_empty_element(node):
            remove_node(node)

        # merge text in span or strong to parent p tag
        if node.tag.lower() == 'p':
            etree.strip_tags(node, 'span')
            etree.strip_tags(node, 'strong')

        # if a div tag does not contain any sub node, it could be converted to p node.
        if node.tag.lower() == 'div' and not node.getchildren():
            node.tag = 'p'

        if node.tag.lower() == 'span' and not node.getchildren():
            node.tag = 'p'

        # remove empty p tag
        if node.tag.lower() == 'p' and not node.xpath('.//img'):
            if not (node.text and node.text.strip()):
                drop_tag(node)

        class_name = node.get('class')
        if class_name:
            if class_name in USELESS_ATTR:
                remove_node(node)
                break


def html2element(html):
    html = re.sub('</?br.*?>', '', html)
    element = fromstring(html)
    return element


def pre_parse(element):
    normalize_node(element)

    return element


def remove_noise_node(element, noise_xpath_list):
    noise_xpath_list = noise_xpath_list or config.get('noise_node_list')
    if not noise_xpath_list:
        return
    for noise_xpath in noise_xpath_list:
        nodes = element.xpath(noise_xpath)
        for node in nodes:
            remove_node(node)
    return element


def iter_node(element: HtmlElement):
    yield element
    for sub_element in element:
        if isinstance(sub_element, HtmlElement):
            yield from iter_node(sub_element)


def remove_node(node: HtmlElement):
    """
    this is a in-place operation, not necessary to return
    :param node:
    :return:
    """
    parent = node.getparent()
    if parent is not None:
        parent.remove(node)


def drop_tag(node: HtmlElement):
    """
    only delete the tag, but merge its text to parent.
    :param node:
    :return:
    """
    parent = node.getparent()
    if parent is not None:
        node.drop_tag()


def is_empty_element(node: HtmlElement):
    return not node.getchildren() and not node.text


def pad_host_for_images(host, url):
    """
    网站上的图片可能有如下几种格式：

    完整的绝对路径：https://xxx.com/1.jpg
    完全不含 host 的相对路径： /1.jpg
    含 host 但是不含 scheme:  xxx.com/1.jpg 或者  ://xxx.com/1.jpg

    :param host:
    :param url:
    :return:
    """
    if url.startswith('http'):
        return url
    parsed_uri = urlparse(host)
    scheme = parsed_uri.scheme
    if url.startswith(':'):
        return f'{scheme}{url}'
    if url.startswith('//'):
        return f'{scheme}:{url}'
    return urljoin(host, url)


def read_config():
    if os.path.exists('.xgne'):
        with open('.xgne', encoding='utf-8') as f:
            config_text = f.read()
        config = yaml.safe_load(config_text)
        return config
    return {}


def get_high_weight_keyword_pattern():
    return re.compile('|'.join(HIGH_WEIGHT_ARRT_KEYWORD), flags=re.I)


def get_longest_common_sub_string(str1: str, str2: str) -> str:
    """
    获取两个字符串的最长公共子串。

    构造一个矩阵，横向是字符串1，纵向是字符串2，例如：

      BL是天才！？
    听0 0 0 0 00 0
    说0 0 0 0 00 0
    B1 0 0 0 00 0
    L0 1 0 0 00 0
    是0 0 1 0 00 0
    天0 0 0 1 00 0
    才0 0 0 0 10 0
    ！0 0 0 0 01 0

    显然，只要斜对角线最长的就是最长公共子串

    :param str1: 第一个字符串
    :param str2: 第二个字符串
    :return: 最长公共子串
    """
    if not all([str1, str2]):
        return ''

    # 构造一个矩阵，横向是字符串1，纵向是字符串2
    matrix = [[0] * (len(str2) + 1) for _ in range(len(str1) + 1)]
    max_length = 0
    start_position = 0

    for index_of_str1 in range(1, len(str1) + 1):
        for index_of_str2 in range(1, len(str2) + 1):
            if str1[index_of_str1 - 1] == str2[index_of_str2 - 1]:
                matrix[index_of_str1][index_of_str2] = matrix[index_of_str1 - 1][index_of_str2 - 1] + 1
                if matrix[index_of_str1][index_of_str2] > max_length:
                    max_length = matrix[index_of_str1][index_of_str2]
                    start_position = index_of_str1 - max_length
            else:
                matrix[index_of_str1][index_of_str2] = 0

    return str1[start_position: start_position + max_length]


def normalize_text(html):
    """
    使用 NFKC 对网页源代码进行归一化，把特殊符号转换为普通符号
    :param html:
    :return:
    """
    return unicodedata.normalize('NFKC', html)


def urljoin_if_valid(base_url: str, url: str) -> str:
    """拼接基本URL和可能是相对URL的地址，防止由于解析导致的无效URL。

    Args:
        base_url (str): 基本URL（即文章的URL）
        url (str): 相对或绝对URL

    Returns:
        str: 如果有效则返回拼接的URL，否则返回空字符串
    """

    try:
        res = urljoin(base_url, url)
        return res
    except ValueError:
        return ""


def url_to_filetype(abs_url: str) -> None:
    """
    输入一个URL，输出由URL指定的文件的文件类型。对于没有文件类型的情况，返回None。
    'http://blahblah/images/car.jpg' -> 'jpg'
    'http://yahoo.com'               -> None
    """
    path = urlparse(abs_url).path.rstrip('/')
    path_chunks = [x for x in path.split("/") if x]
    last_chunk = path_chunks[-1].split('.')  # 最后一块通常是文件
    if len(last_chunk) < 2:
        return None
    file_type = last_chunk[-1]
    # 假设文件扩展名最多为5个字符长
    return file_type.lower() if len(file_type) <= 5 or file_type.lower() in ALLOWED_TYPES else None


def get_requests_params():
    '''
    获取请求头的参数
    :return:
    '''
    return {
        "timeout": 7,
        "proxies": {},
        "headers": {
            "User-Agent": f"xgne/{__version__}",
        },
        "cookies": cj(),
    }


# 通用文本过滤
def universal_filter(text):
    if text:
        # \u1234
        text = text.replace(u'\u200b', '')
        text = text.replace(u'\u2002', '')
        text = text.replace(u'\u3000', '')
        text = text.replace(u'\ufeff', '')
        # \x??
        text = text.replace(u'\xa0', '')
        text = text.replace(u'\x7f', '')
        # &??
        text = text.replace('&nbsp', ' ')
        text = text.replace('&ldquo;', '"')
        text = text.replace('&rdquo;', '"')
        text = text.replace('&bull;', '•')
        text = text.replace('&mdash;', '—')
        text = text.replace('&lsquo;', "'")
        text = text.replace('&rsquo;', "'")
        text = text.replace('&hellip;', '…')
        text = text.replace('&middot;', '·')
        text = text.replace('&quot;', '"')
        text = text.replace('&amp;', '&')
        text = text.replace('&#39;', "'")
        text = text.replace('&deg;', "°")
        text = text.replace('&times;', "×")
        text = text.replace('&beta;', "β")
        text = text.replace('&ndash;', "–")
        # \n, \r, \t
        text = text.replace('\n', '')
        text = text.replace('\r', '')
        text = text.replace('\t', '')
        # '  ?  '
        text = text.strip()
    return text


class FileHelper(object):
    @staticmethod
    def loadResourceFile(filename):
        if not os.path.isabs(filename):
            dirpath = os.path.abspath(os.path.dirname(__file__))
            path = os.path.join(dirpath, 'resources', filename)
        else:
            path = filename
        try:
            f = codecs.open(path, 'r', 'utf-8')
            content = f.read()
            f.close()
            return content
        except IOError:
            raise IOError("Couldn't open file %s" % path)


config = read_config()



import re
import jieba.posseg as pseg


class InformationExtractor:
    def __init__(self):
        # 更宽泛的电话号码匹配规则
        self.phone_pattern = re.compile(r'\b(?:\+?\d{1,3}[-.●]?)?\(?(?:\d{1,4}\)?[-.●]?)?\d{6,}\b')

    def extract_phone_numbers(self, text):
        # 从文本中提取更宽泛的手机号
        return re.findall(self.phone_pattern, text)

    def extract_locations(self, text):
        # 使用结巴分词和词性标注提取地名
        locations = []
        words = pseg.cut(text)
        for word, flag in words:
            if flag == 'ns':  # 'ns' 表示地名
                locations.append(word)
        return locations



import re

from lxml.html import HtmlElement, Element
from collections import Counter

from ..defaults import META_IMAGE_TAGS
from ..parsers import Parser


class HeadMetaExtractor:
    def __init__(self):
        # 定义用于匹配 URL 的正则表达式
        # self.url_pattern = re.compile(r'https?://\S+')
        self.url_pattern = re.compile(r'https?://[^\s/]+')
        self.parser = Parser()

    def extract_host(self, head_meta):
        # 提取所有可能的 host
        potential_hosts = []
        for key, value in head_meta.items():
            # 使用正则表达式检查值是否是有效的 URL
            matches = self.url_pattern.findall(value)
            for match in matches:
                potential_hosts.append(match)

        # 如果有有效的 host，选择出现次数最多的作为最终的 host
        if potential_hosts:
            most_common_host = Counter(potential_hosts).most_common(1)[0][0]
            return most_common_host
        else:
            return None

    def extractor(self, element: HtmlElement):
        # 选择包含 meta 元素的标签
        meta_tags = element.xpath('//head/meta')

        # 遍历每个 meta 元素
        head_meta = {}
        for tag in meta_tags:
            # 获取 meta 元素的属性
            property_value = tag.get('property')
            name_value = tag.get('name')
            content_value = tag.get('content')

            # 构建字典
            if property_value and content_value:
                head_meta[property_value] = content_value
            elif name_value and content_value:
                head_meta[name_value] = content_value
        # 提取 host
        host = self.extract_host(head_meta)

        # 将 host 添加到 head_meta 中
        head_meta['host'] = host
        head_meta['top_image'] = self._get_meta_image(element)

        return head_meta

    def _get_meta_image(self, element: Element) -> str:
        """从文档中获取可能的图片标签。"""
        candidates = []
        for elem in META_IMAGE_TAGS:
            if elem["tag"] == "meta":
                candidates.append(
                    (self._get_meta_field(element, elem["field"]), elem["score"])
                )
            else:
                img = self.parser.getElementsByTag(
                    element,
                    attr=elem["attr"],
                    value=elem["value"],
                    use_regex=("|" in elem["value"]),
                )
                if img:
                    candidates.append((img[0].get("href"), elem["score"]))
        candidates = [c for c in candidates if c[0]]

        candidates.sort(key=lambda x: x[1], reverse=True)

        return candidates[0][0] if candidates else ""

    def _get_meta_field(self, element: Element, field: str) -> str:
        """从文档中提取给定的 meta 字段。"""
        metafield = self.parser.css_select(element, field)
        if metafield:
            return metafield[0].get("content", "").strip()
        return ""



import logging
import urllib.parse
from copy import copy
import re
from typing import List, Optional
from PIL import Image, ImageFile
import requests
from lxml.html import Element

from ..defaults import META_IMAGE_TAGS, TOP_IMAGE_SETTINGS
from ..parsers import Parser
from ..utils import urljoin_if_valid, url_to_filetype, get_requests_params

log = logging.getLogger(__name__)


class ImageExtractor:
    """文章中的图片提取器，用于获取顶部图片、图片列表、网站图标等。"""

    def __init__(self):
        self.parser = Parser()
        self.top_image: Optional[str] = None
        self.meta_image: Optional[str] = None
        self.images: List[str] = []
        self.favicon: Optional[str] = None
        self._chunksize = 1024

    def extractor(
            self, doc: Element, top_node: Element, article_url: str
    ) -> None:
        """从文档中提取图片的主要方法。

        Args:
            doc (Element): 文档元素
            top_node (Element): 顶部节点
            article_url (str): 文章链接
        """
        self.favicon = self._get_favicon(doc)

        self.meta_image = self._get_meta_image(doc)
        if self.meta_image:
            self.meta_image = urljoin_if_valid(article_url, self.meta_image)
        self.images = [
            urljoin_if_valid(article_url, u)
            for u in self._get_images(doc)
            if u and u.strip()
        ]
        self.top_image = self._get_top_image(doc, top_node, article_url)

    def _get_favicon(self, doc: Element) -> str:
        """从网站中提取图标，参考 http://en.wikipedia.org/wiki/Favicon
        <link rel="shortcut icon" type="image/png" href="favicon.png" />
        <link rel="icon" type="image/png" href="favicon.png" />
        """
        kwargs = {"tag": "link", "attr": "rel", "value": "icon"}
        meta = self.parser.getElementsByTag(doc, **kwargs)
        if meta:
            favicon = self.parser.getAttribute(meta[0], "href")
            return favicon
        return ""

    def _get_meta_field(self, doc: Element, field: str) -> str:
        """从文档中提取给定的 meta 字段。"""
        metafield = self.parser.css_select(doc, field)
        if metafield:
            return metafield[0].get("content", "").strip()
        return ""

    def _get_meta_image(self, doc: Element) -> str:
        """从文档中获取可能的图片标签。"""
        candidates = []
        for elem in META_IMAGE_TAGS:
            if elem["tag"] == "meta":
                candidates.append(
                    (self._get_meta_field(doc, elem["field"]), elem["score"])
                )
            else:
                img = self.parser.getElementsByTag(
                    doc,
                    attr=elem["attr"],
                    value=elem["value"],
                    use_regex=("|" in elem["value"]),
                )
                if img:
                    candidates.append((img[0].get("href"), elem["score"]))
        candidates = [c for c in candidates if c[0]]

        candidates.sort(key=lambda x: x[1], reverse=True)

        return candidates[0][0] if candidates else ""

    def _get_images(self, doc: Element) -> List[str]:
        """从文档中获取所有图片的链接。"""
        images = [
            x.get("src")
            for x in self.parser.getElementsByTag(doc, tag="img")
            if x.get("src")
        ]

        return images

    def _get_top_image(
            self, doc: Element, top_node: Element, article_url: str
    ) -> str:
        """获取顶部图片链接。"""

        def node_distance(node1, node2):
            path1 = node1.getroottree().getpath(node1).split("/")
            path2 = node2.getroottree().getpath(node2).split("/")
            for i, (step1, step2) in enumerate(zip(path1, path2)):
                if step1 != step2:
                    return len(path1[i:]) + len(path2[i:])

            return abs(len(path1) - len(path2))

        if self.meta_image:
            if self._check_image_size(
                    self.meta_image, article_url
            ):
                return self.meta_image

        img_cand = []
        for img in self.parser.getElementsByTag(doc, tag="img"):
            if not img.get("src"):
                continue
            if img.get("src").startswith("data:"):
                continue

            if top_node is not None:
                distance = node_distance(top_node, img)
                img_cand.append((img, distance))
            else:
                if self._check_image_size(img.get("src"), article_url):
                    return img.get("src")

        img_cand.sort(key=lambda x: x[1])

        for img in img_cand:
            if self._check_image_size(img[0].get("src"), article_url):
                return img[0].get("src")

        return ""

    def _check_image_size(self, url: str, referer: Optional[str]) -> bool:
        """检查图片的尺寸是否符合设定的要求。"""
        img = self._fetch_image(
            url,
            referer,
        )
        if not img:
            return False

        width, height = img.size

        if TOP_IMAGE_SETTINGS["min_width"] > width:
            return False
        if TOP_IMAGE_SETTINGS["min_height"] > height:
            return False
        if TOP_IMAGE_SETTINGS["min_area"] > width * height:
            return False

        if (
                re.search(r"(logo|sprite)", url, re.IGNORECASE)
                and TOP_IMAGE_SETTINGS["min_area"] > width * height / 10
        ):
            return False

        return True

    def _fetch_image(self, url: str, referer: Optional[str]) -> Optional[Image.Image]:
        """从给定的 URL 获取图片，并检查是否符合要求。"""

        def clean_url(url):
            """将 URL 中的 Unicode 数据进行编码"""
            if not isinstance(url, str):
                return url

            url = url.encode("utf8")
            url = "".join(
                [
                    urllib.parse.quote(c) if ord(c) >= 127 else c
                    for c in url.decode("utf-8")
                ]
            )
            return url

        requests_params = copy(get_requests_params())
        requests_params["headers"]["Referer"] = referer
        max_retries = TOP_IMAGE_SETTINGS["max_retries"]

        cur_try = 0
        url = clean_url(url)
        if not url or not url.startswith(("http://", "https://")):
            return None

        response = None
        while True:
            try:
                response = requests.get(
                    url,
                    stream=True,
                    **requests_params,
                )

                content_type = response.headers.get("Content-Type")

                if not content_type or "image" not in content_type.lower():
                    return None

                p = ImageFile.Parser()
                new_data = response.raw.read(self._chunksize)
                while not p.image and new_data:
                    try:
                        p.feed(new_data)
                    except (IOError, ValueError) as e:
                        log.warning(
                            "在获取图片时发生错误 %s: %s 来源: %s",
                            str(e),
                            url,
                            requests_params["headers"].get("Referer"),
                        )
                        return None
                    except Exception as e:
                        # 对于一些 favicon.ico 图片，图片很小，
                        # 导致我们的 PIL feed() 方法无法通过长度测试。
                        is_favicon = url_to_filetype(url) == "ico"
                        if not is_favicon:
                            raise e
                        return None
                    new_data = response.raw.read(self._chunksize)
                return p.image
            except requests.exceptions.RequestException:
                cur_try += 1
                if cur_try >= max_retries:
                    log.warning(
                        "在获取图片时发生错误: %s 来源: %s",
                        url,
                        requests_params["headers"].get("Referer"),
                    )
                    return None
            finally:
                if response is not None:
                    response.raw.close()
                    if response.raw._connection:
                        response.raw._connection.close()
